/*
 *  Program to try and read Tops-20 Dumper format tapes
 *
 *                   Jim Guyton,  Rand Corp.
 *                   Version 1 (10/20/82)
 *                   jdg:                  -n added 6/11/83
 *                   pell@lysator.liu.se:  -b added 90-01-14
 *		     also fixed correct generation of unix file names,
 *		     with automatic directory generation.
 *		     pell: fixed a bug in the above 92-11-24.
 * Sorry, Jim, but after struggling with the IMO unorthodox brace style
 * and curious indentation, I just reformatted the source.  /Pell
 */

/* TODO: fix extraction of binaries (36-bit) */

#include <stdio.h>
#include <sys/file.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include "dump.h"
/* logfile should be changable */
#define LOGFILE "Logfile"
#define MAX_BLOCK_FAC 20	/* Max blocking factor /Pell */

char *ctime(), *index(), *rindex();
char *unixname(), *unixdir();

int  fdTape;                    /* File handle for Dumper-20 format tape */
char tapeblock[MAX_BLOCK_FAC][TAPEBLK];	/* One logical record from tape */
FILE *fpFile;                   /* Output file handle on extracts */
int debug = 0;
int textflg = 0;                /* Non-zero if retr binary files as text */
int numflg = 0;                 /* Non-zero if using numeric filenames */
int number;                     /* Current output file "number" */

#define TAPE "/dev/rmt0"        /* Default input tape */

int  bytesize;          /* Number of bits/byte in current file */
int  numbytes;          /* Number of bytes in current file */
int  pgcount;           /* Number of twenex pages in file */
int  blfac;		/* Current blocking factor /Pell */

char *pattern = 0;      /* Filename match pattern */

/*
  pgm  [-f tapefile] [-b blocking-factor] [-t] [-n number] pattern
  
  no pattern == directory only
  no tapefile == /dev/rmt0
  -b == Set other blocking factor than 1
  -t == try to pretend files are 7-bit ascii
  -n == use numeric filenames in extracts, number is 1st name
  */

main(argc, argv)
     int argc;
     char *argv[];
{
  char *tape = TAPE;              /* Pathname for tape device/file */
  int rc;
  int b;
  
  blfac = 1;		/* Default blocking factor */
  /* Do switch parsing */
  
  while(argc>1 && argv[1][0] == '-') {
    switch(argv[1][1]) {
    case 'f':
      if (argc <= 2) punt("Need filename after -f\n");
      tape = argv[2];
      argc--; argv++;
      break;
    case 'b':
      if (argc <= 2) punt("Need blocking factor after -b\n");
      blfac = atoi(argv[2]);
      if (blfac > MAX_BLOCK_FAC) {
	punt("Blocking factor must be less than %d!\n",
	     MAX_BLOCK_FAC);
	exit(1);
      }
      argc--; argv++;
      break;
    case 't':             /* Force text mode on "binary" files */
      textflg = 1;
      break;
    case 'd':
      debug = atoi(&argv[1][2]);
      printf("Debug value set to %d\n", debug);
      break;
    case 'n':               /* numeric output filenames */
      if (argc <= 2) punt("Need number after -n\n");
      number = atoi(argv[2]);         /* First file name */
      numflg = 1;
      argc--; argv++;
      break;
    default:
      printf("unknown flag %s\n", argv[1]);
      exit(1);
      break;
    }
    argc--;  argv++;
  }
  
  
  if (argc > 1)
    pattern = argv[1];
  
  fdTape = open(tape, 0);	/* Open tape for read */
  if (fdTape == -1)
    punt("Couldn't open 'tape' file %s\n", tape);
  
  for ( ; ; ) {			/* Loop till end of tape */
    /*** Read a block ***/
    rc = read(fdTape, tapeblock, blfac*TAPEBLK);
    if (rc != blfac*TAPEBLK) {
      if (rc != 0)
	punt("Oops.  Read block len=%d\n", rc);

      printf("\nEnd of tape.\n");
      exit(0);        /* Normal exit */
    }
    for (b=0; b<blfac; b++) {
      /*** Do something with it ***/
      switch(getrecordtype(tapeblock[b]))
	{
	case RectypeData:             /* Data block */
	  doDatablock(tapeblock[b]);
	  break;
	  
	case RectypeTphd:             /* Saveset header */
	  doSaveset(tapeblock[b], 0);
	  break;
	  
	case RectypeFlhd:             /* File header */
	  doFileHeader(tapeblock[b]);
	  break;
	  
	case RectypeFltr:             /* File trailer */
	  doFileTrailer(tapeblock[b]);
	  break;
	  
	case RectypeTptr:             /* Tape trailer */
	  doTapeTrailer(tapeblock[b]);
	  break;
	  
	case RectypeUsr:              /* User directory info ? */
	  printf("User info record skipped\n");
	  break;
	  
	case RectypeCtph:             /* Continued saveset hdr */
	  doSaveset(tapeblock[b], 1);
	  break;
	  
	case RectypeFill:             /* Fill record */
	  printf("Fill record skipped\n");
	  break;
	  
	default:
	  punt("Unknown record type 0x%x\n",
	       getrecordtype(tapeblock[b]));
	  break;
	}
    }
  }
}

/* Get the "record type" from the tape block header.  Since it */
/* is stored in 2's complement form, negate it before returning */

getrecordtype(block)
     char *block;
{
  long int tl;
  tl = getfield(block, WdoffRectype, BtoffRectype, BtlenRectype);
  return( (int) -tl);
}

int masks[32] = {		/* bitmasks for different length fields */
  0x00000001, 0x00000003, 0x00000007,
  0x0000000f, 0x0000001f, 0x0000003f, 0x0000007f,
  0x000000ff, 0x000001ff, 0x000003ff, 0x000007ff,
  0x00000fff, 0x00001fff, 0x00003fff, 0x00007fff,
  0x0000ffff, 0x0001ffff, 0x0003ffff, 0x0007ffff,
  0x000fffff, 0x001fffff, 0x003fffff, 0x007fffff,
  0x00ffffff, 0x01ffffff, 0x03ffffff, 0x07ffffff,
  0x0fffffff, 0x1fffffff, 0x3fffffff, 0x7fffffff,
  0xffffffff
};


long getfield(block, wordoff, bitoff, bitlen)
     char *block;            /* Tape block record */
     int wordoff;            /* 36-bit word offset */
     int bitoff;             /* Bit offset of field (from msb) */
     int bitlen;             /* Bit length of field */
{
  char *p;                /* Used to point into record */
  long int w32;           /* First 32 bits of the 36 bit word */
  int   w4;               /* Last 4 bits of the 36 bit word */
  long  w = 0;            /* the word to return */
  
  
  /* First, the "illegal" kludge */
  if (bitoff == 0 && bitlen == 36) {
    bitoff = 4;
    bitlen = 32;
  }
  if (bitlen > 32) punt("I can't get that large a field!\n");
  
  /* A PDP-10 (or 20) 36-bit word is laid out with the first 32 bits
     as the first 4 bytes and the last 4 bits are the low order 4 bits
     of the 5th byte.   The high 4 bits of that byte should be zero */
  
  p = block + (5*wordoff);        /* Get ptr to word of interest */
  w32 = *p++ & 0377;                      /* First byte */
  w32 = (w32 << 8) | (*p++ & 0377);       /* 2nd */
  w32 = (w32 << 8) | (*p++ & 0377);       /* 3rd */
  w32 = (w32 << 8) | (*p++ & 0377);       /* 4th */
  w4  = *p;                               /* 5th */
  if (w4 > 017) punt("Not a PDP-10 tape!  w4=%o\n", w4);
  
  
  /* Get the field right justified in the word "w".
     There are three cases that I have to handle:
     [1] field is contained in w32
     [2] field crosses w32 and w4
     [3] field is contained in w4
     */
  
  if (bitoff+bitlen <= 32) {	/* [1] field is contained in w32 */
    w = w32 >> (32 - (bitoff+bitlen));
  }
  else if (bitoff <= 32) {	/* [2] field crosses boundary */
    w =  (w32 << (bitoff+bitlen-32))
      | (w4  >> (36 - (bitoff+bitlen)));
  }
  else {			/* [3] field is contained in w4 */
    w = w4 >> (36 - (bitoff+bitlen));
  }
  w = w & masks[bitlen-1];	/* Trim to proper size */
  return(w);
}

doDatablock(block)
     char *block;
{
  static char buf[(512*5)+1];	/* A page of characters */
  int ct;
  if (debug > 10) printf("*");
  if (fpFile == NULL) return;
  /* 7 bit ascii only for now */
  if (numbytes > 512*5) ct = 512*5;
  else ct = numbytes;
  
  getstring(block, buf, 6, ct);
  buf[ct] = 0;
  /* Pell: replaced fprintf(fpFile, "%s", buf); with fputs() */
  /* Should be fwrite() to allow for 0's, but binary extraction doesn't */
  /* work at the moment anyway. */
  fputs(buf, fpFile);
  numbytes -= ct;
}

doSaveset(block, contflag)
     char *block;
     int  contflag;
{
  static char name[100];
  long t;
  
  if (debug > 10) printf("\nSaveset header:");
  getstring(block, name, WdoffSSName, sizeof(name));
  
  t = unixtime(block, WdoffSSDate);
  printf("Saveset '%s', %s\n", name, ctime(&t));
  
}

doFileHeader(block)
     char *block;
{
  /* The constants should be taken from system include files. */
  static char name[200], fname[250];	/* Pell */
  static char olddir[100];
  long t;			/* The time in unix format */
  char *ts, *dir;
  
  if (debug > 5) printf("File Header block:\n");
  
  getstring(block, name, WdoffFLName, sizeof(name));
  ts = index(name, ';');	/* Chop off ;Pprotection;Aacct */
  *ts = 0;
  
  t = unixtime(block, WdoffFDB_Wrt);
  ts = ctime(&t) + 4;		/* Skip over day-name field */
  ts[strlen(ts)-1] = 0;		/* Chop off \n at end */
  
  bytesize = getfield(block, WdoffFDB_BSZ, BtoffFDB_BSZ, BtlenFDB_BSZ);
  numbytes = getfield(block, WdoffFDB_Size, BtoffFDB_Size, BtlenFDB_Size);
  pgcount  = getfield(block, WdoffFDB_PGC, BtoffFDB_PGC, BtlenFDB_PGC);
  
  printf("%6d %11d(%2d) %s %s\n", pgcount, numbytes, bytesize, ts, name);
  
  if (pattern && match(name, pattern)) {
    /* Special hack for bad files */
    if (bytesize != 7 && textflg) {
      if (bytesize == 0 || bytesize == 36) { /* Sigh */
	bytesize = 7;
	numbytes = numbytes * 5;
      }
    }
    if (bytesize != 7)
      fprintf(stderr, "Skipping -- binary file.\n");
    else {
      dir = unixdir(name);	/* Pell */
      if (strcmp(dir, olddir)) { /* Different. Maybe try to create */
	if (mk_unix_dir(dir))
	  punt("Couldn't create directory %s!\n", dir);
	strcpy(olddir, dir); /* Update current dir */
      }
      sprintf(fname, "%s/%s", dir, unixname(name));
      fpFile = fopen(fname, "w");
      if (fpFile == NULL)
	punt("Can't open %s for write!\n", fname);
      printf("Extracting\n");
    }
  }
  else
    fpFile = NULL;
}

doFileTrailer(block)
     char *block;
{
  if (debug > 10) printf(" File trailer\n");
  if (fpFile != NULL) {
    fclose(fpFile);
    fpFile = NULL;
  }
}

doTapeTrailer(block)
     char *block;
{
  if (debug > 10) printf("Tape Trailer");
}

punt(s, arg)
     char *s;
     int  arg;
{
  fprintf(stderr, s, arg);
  /* This perror should only be done if there's a system error. /Pell */
  perror("read20: ");
  exit(1);
}

getstring(block, s, wordoff, max)
     char *block;		/* Tape block */
     char *s;			/* Destination string buffer */
     int  wordoff;		/* 36-bit offset from start of tape block */
     int  max;			/* Max number of characters to xfer into s */
{
  int i;			/* Counter for five characters per word */
  int ct = 0;			/* Number of characters loaded so far */
  char *orig = s;		/* Save for debugging */
  
  while (ct < max) {
    for (i = 0; i < 5; i++) {
      *s = getfield(block, wordoff, i*7, 7);
      if (*s == 0) return;
      s++;
    }
    wordoff++;
    ct += 5;
  }
  /**     punt("String greater than %d characters.\n", max);   **/
}

#define SecPerTick  (24.*60.*60.)/0777777
#define DayBaseDelta 0117213            /* Unix day 0 in Tenex format */
   
long unixtime(block, wordoff)
     char *block;
     int  wordoff;
{
  long int t, s;
  
  t = getfield(block, wordoff, 0, 18);    /* First half is day */
  t -= DayBaseDelta;                      /* Switch to unix base */
  /* Now has # days since */
  /* Jan 1, 1970 */
  
  s = getfield(block, wordoff, 18, 18);   /* 2nd half is fraction day */
  s = s * SecPerTick;                     /* Turn into seconds */
  
  s += t*24*60*60;                        /* Add day base */
  return(s);
}


/* See if pattern is in name (very simple) */
match(name, pattern)
     char *name, *pattern;
{
  int  plen = strlen(pattern);
  
  while ((name=index(name, *pattern))) {
    if (strncmp(name, pattern, plen)==0) return(1);
    name++;
    if (*name == 0) return(0);         /* May not need */
  }
  return(0);
}

char *unixdir(name)		/* Pell */
     char *name;
{
  static char dir[200];
  char *t;
  int i;
  
  name = index(name, '<');	/* Trim off device */
  t = rindex(name, '>');	/* find end of directory */
  
  strncpy(dir, name+1, t-name-1);
  dir[t-name-1] = 0;		/* Terminate */
  lowify(dir);

  for (i=0; i<t-name-1; i++)
    if (dir[i] == '.') dir[i] = '/'; /* Unixify */

  return(dir);
}

int mk_unix_dir(dir)
     char *dir;
{
  /* Go through, and try to create each directory in the path */
  char full_path[2048];
  char *part;
  char *dirs=strdup(dir);
  
  if (getwd(full_path) == 0)
    punt("getwd failed!\n");
  
  part = strtok(dirs, "/");
  while (part != 0) {
    strcat(full_path, "/");
    strcat(full_path, part);
    if (access(full_path, W_OK|X_OK) < 0) {	/* Try to create */
      if (mkdir(full_path, 0777) < 0) { /* Didn't work */
	return -1;
      }
    }
    part = strtok((char *)0, "/");
  }				/* while */
  free(dirs);
  return 0;			/* all ok */
}

char *unixname(name)
     char *name;
{
  static char newname[200];
  static FILE *log = NULL;
  char *t;
  
  if (numflg) {			/* If numeric filenames */
    if (log == NULL) log = fopen(LOGFILE, "a");
    fprintf(log, "%d is %s\n", number, name);
    sprintf(newname, "%d", number++);
    return(newname);
  }
  
  name = index(name, '<');	/* Trim off device */
  /* I know I should check for ^V... */
  t = rindex(name, '>');	/* find end of directory */
  
  strcpy(newname, ++t);		/* Skip over the > */
  lowify(newname);
  return(newname);
}

lowify(str)			/* Pell */
     char *str;
{
  while (str && *str) {
    if (isascii(*str) && isupper(*str)) *str = tolower(*str);
    str++;
  }
}
